"use strict";

var _interopRequireDefault = require("@babel/runtime/helpers/interopRequireDefault").default;
Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.createEventHandler = exports.Operations = void 0;
var _error = _interopRequireDefault(require("@webiny/error"));
var _clientDynamodb = require("@webiny/aws-sdk/client-dynamodb");
var _apiElasticsearch = require("@webiny/api-elasticsearch");
var _handlerAws = require("@webiny/handler-aws");
var _types = require("@webiny/api-elasticsearch/operations/types");
var _pRetry = _interopRequireDefault(require("p-retry"));
var _NotEnoughRemainingTimeError = require("./NotEnoughRemainingTimeError");
let Operations = /*#__PURE__*/function (Operations) {
  Operations["INSERT"] = "INSERT";
  Operations["MODIFY"] = "MODIFY";
  Operations["REMOVE"] = "REMOVE";
  return Operations;
}({});
exports.Operations = Operations;
const getError = item => {
  if (!item.index?.error?.reason) {
    return null;
  }
  const reason = item.index.error.reason;
  if (reason.match(/no such index \[([a-zA-Z0-9_-]+)\]/) !== null) {
    return "index";
  }
  return reason;
};
const checkErrors = result => {
  if (!result || !result.body || !result.body.items) {
    return;
  }
  for (const item of result.body.items) {
    const err = getError(item);
    if (!err) {
      continue;
    } else if (err === "index") {
      if (process.env.DEBUG === "true") {
        console.error("Bulk response", JSON.stringify(result, null, 2));
      }
      continue;
    }
    console.error(item.error);
    throw new _error.default(err, "DYNAMODB_TO_ELASTICSEARCH_ERROR", item);
  }
};
const getNumberEnvVariable = (name, def) => {
  const input = process.env[name];
  const value = Number(input);
  if (isNaN(value)) {
    return def;
  } else if (value <= 0) {
    return def;
  }
  return value;
};
const MAX_PROCESSOR_PERCENT = getNumberEnvVariable("MAX_ES_PROCESSOR", process.env.NODE_ENV === "test" ? 101 : 98);
const unmarshall = value => {
  if (!value) {
    return undefined;
  }
  return (0, _clientDynamodb.unmarshall)(value);
};
const minRemainingSecondsToTimeout = 120;

/**
 * Also, we need to set the maximum running time for the Lambda Function.
 * https://github.com/webiny/webiny-js/blob/f7352d418da2b5ae0b781376be46785aa7ac6ae0/packages/pulumi-aws/src/apps/core/CoreOpenSearch.ts#L232
 * https://github.com/webiny/webiny-js/blob/f7352d418da2b5ae0b781376be46785aa7ac6ae0/packages/pulumi-aws/src/apps/core/CoreElasticSearch.ts#L218
 */
const MAX_RUNNING_TIME = 900;
const createEventHandler = () => {
  return (0, _handlerAws.createDynamoDBEventHandler)(async ({
    event,
    context: ctx,
    lambdaContext
  }) => {
    const timer = (0, _handlerAws.timerFactory)(lambdaContext);
    const context = ctx;
    if (!context.elasticsearch) {
      console.error("Missing elasticsearch definition on context.");
      return null;
    }
    const operations = [];
    const operationIdList = [];
    for (const record of event.Records) {
      const dynamodb = record.dynamodb;
      if (!dynamodb) {
        continue;
      }
      /**
       * TODO: figure out correct types
       */
      // @ts-expect-error
      const newImage = unmarshall(dynamodb.NewImage);

      // Note that with the `REMOVE` event, there is no `NewImage` property. Which means,
      // if the `newImage` is `undefined`, we are dealing with a `REMOVE` event and we still
      // need to process it.
      if (newImage && newImage.ignore === true) {
        continue;
      }
      /**
       * TODO: figure out correct types
       */
      // @ts-expect-error
      const keys = unmarshall(dynamodb.Keys);
      if (!keys?.PK || !keys.SK) {
        continue;
      }
      const _id = `${keys.PK}:${keys.SK}`;
      /**
       * TODO: figure out correct types
       */
      // @ts-expect-error
      const oldImage = unmarshall(dynamodb.OldImage);
      const operation = record.eventName;

      /**
       * On operations other than REMOVE we decompress the data and store it into the Elasticsearch.
       * No need to try to decompress if operation is REMOVE since there is no data sent into that operation.
       */
      let data = undefined;
      if (newImage && operation !== Operations.REMOVE) {
        /**
         * We must decompress the data that is going into the Elasticsearch.
         */
        data = await (0, _apiElasticsearch.decompress)(context.plugins, newImage.data);
        /**
         * No point in writing null or undefined data into the Elasticsearch.
         * This might happen on some error while decompressing. We will log it.
         *
         * Data should NEVER be null or undefined in the Elasticsearch DynamoDB table, unless it is a delete operations.
         * If it is - it is a bug.
         */
        if (data === undefined || data === null) {
          console.error(`Could not get decompressed data, skipping ES operation "${operation}", ID ${_id}`);
          continue;
        }
      }
      operationIdList.push(_id);
      switch (record.eventName) {
        case Operations.INSERT:
        case Operations.MODIFY:
          if (newImage) {
            operations.push({
              index: {
                _id,
                _index: newImage.index
              }
            }, data);
          }
          break;
        case Operations.REMOVE:
          operations.push({
            delete: {
              _id,
              _index: oldImage?.index || "unknown"
            }
          });
          break;
        default:
          break;
      }
    }
    /**
     * No need to do anything if there are no operations.
     */
    if (operations.length === 0) {
      return null;
    }
    /**
     * Wrap the code we need to run into the function, so it can be called within itself.
     */
    const execute = async () => {
      const remainingTime = timer.getRemainingSeconds();
      const runningTime = MAX_RUNNING_TIME - remainingTime;
      const maxWaitingTime = remainingTime - 90;
      if (process.env.DEBUG === "true") {
        console.debug(`The Lambda is already running for ${runningTime}s. Setting Health Check max waiting time: ${maxWaitingTime}s`);
      }
      const healthCheck = (0, _apiElasticsearch.createWaitUntilHealthy)(context.elasticsearch, {
        minClusterHealthStatus: _types.ElasticsearchCatClusterHealthStatus.Yellow,
        waitingTimeStep: 30,
        maxProcessorPercent: MAX_PROCESSOR_PERCENT,
        maxWaitingTime
      });
      try {
        await healthCheck.wait({
          async onUnhealthy({
            startedAt,
            runs,
            mustEndAt,
            waitingTimeStep,
            waitingReason
          }) {
            console.debug(`Cluster is unhealthy on run #${runs}.`, {
              startedAt,
              mustEndAt,
              waitingTimeStep,
              waitingReason
            });
          },
          async onTimeout({
            startedAt,
            runs,
            waitingTimeStep,
            mustEndAt,
            waitingReason
          }) {
            console.error(`Cluster health check timeout on run #${runs}.`, {
              startedAt,
              mustEndAt,
              waitingTimeStep,
              waitingReason
            });
          }
        });
      } catch (ex) {
        if (ex instanceof _apiElasticsearch.UnhealthyClusterError || ex instanceof _apiElasticsearch.WaitingHealthyClusterAbortedError) {
          throw ex;
        }
        console.error(`Cluster health check failed.`, ex);
        throw ex;
      }
      try {
        const res = await context.elasticsearch.bulk({
          body: operations
        });
        checkErrors(res);
      } catch (error) {
        if (process.env.DEBUG !== "true") {
          throw error;
        }
        const meta = error?.meta || {};
        delete meta["meta"];
        console.error("Bulk error", JSON.stringify(error, null, 2));
        throw error;
      }
      if (process.env.DEBUG !== "true") {
        return;
      }
      console.info(`Transferred ${operations.length / 2} record operations to Elasticsearch.`);
    };
    const maxRetryTime = getNumberEnvVariable("WEBINY_DYNAMODB_TO_ELASTICSEARCH_MAX_RETRY_TIME", 300000);
    const retries = getNumberEnvVariable("WEBINY_DYNAMODB_TO_ELASTICSEARCH_RETRIES", 20);
    const minTimeout = getNumberEnvVariable("WEBINY_DYNAMODB_TO_ELASTICSEARCH_MIN_TIMEOUT", 1500);
    const maxTimeout = getNumberEnvVariable("WEBINY_DYNAMODB_TO_ELASTICSEARCH_MAX_TIMEOUT", 30000);
    try {
      await (0, _pRetry.default)(execute, {
        maxRetryTime,
        retries,
        minTimeout,
        maxTimeout,
        onFailedAttempt: error => {
          if (timer.getRemainingSeconds() < minRemainingSecondsToTimeout) {
            throw new _NotEnoughRemainingTimeError.NotEnoughRemainingTimeError(error);
          }
          /**
           * We will only log attempts which are after 3/4 of total attempts.
           */
          if (error.attemptNumber < retries * 0.75) {
            return;
          }
          console.error(`Attempt #${error.attemptNumber} failed.`);
          console.error(error);
        }
      });
    } catch (ex) {
      // TODO implement storing of failed operations
      throw ex;
    }
    return null;
  });
};
exports.createEventHandler = createEventHandler;

//# sourceMappingURL=index.js.map